/*
** Der umstlpbare Wrfel nach Paul Schatz
** (C) 1999,2004 Franz Zahaurek <http://www.fzk.at>
**
** Dieses Programm samt Dokumentation ist Freie Software und kann beliebig
** verwendet werden, soferne der obige Copyright-Vermerk nicht entfernt
** wird.
*/
/* $Id: wberechne.c,v 1.11 2004/03/30 17:07:42 fzk Exp $ */

#ifndef OS_WIN32
  #include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wuerfel.h"

void
berechnewg (void) {
    WG *wgp;
    BWINKEL w;

    wg = last = (WG *)NULL;
    

    if (options.dw == 0.) {
	options.dw = 2. * M_PI / 60.;
    }

    /* Lage des Koordinatensystems:

                       | +y
                       |
                       |
                       |
                (0,0,0).----------
                      /        +x
                     /
                    / +z

       w (Winkel) entspricht alpha der Beschreibung. w durchluft die
       Werte 2*PI bis 0 mit einer Schrittweite -dw. Der Startwert 2*PI 
       wird um ein DELTA (= 1/1000) verflscht, um Vielfache von PI zu 
       vermeiden. Fr jede Position wird eine Struktur vom Typ WG
       alloziert, die den Wert w, die Koordinaten der 16 Punkte des
       Wrfelgrtels, sowie die Koordinaten von I und E enthlt. Die
       Indizes haben dabei folgende Bedeutung:

       Index     Bedeutung
       1-8       die obere Hlfte des Wrfelgrtels, Punkte P1 - P8
       9         Evolutionspunkt
       10        Involutionspunkt
       11-18     die unter Hlfte des Wrfelgrtels (Indizes der
                 oberen Hlfte +10). Die Punkte 1,2,5 und 6 sind ident mit 
		 den Punkten 10, 12, 15 und 16
       19        Punkt (2,0,0) auf x-Achse
       20        Punkt (0,2,0) auf y-Achse
       21        Punkt (0,0,2) auf z-Achse

       Nach der Berechnung erfolgt die Prfung, ob die Wrfelkanten ==
       1 und die Scharnierkanten == 1/sqrt(3) lange sind (jeweils fr
       eine Kante).  */

    anzelt = 0;
    /* for(w = 2.*M_PI-DELTA; w > 2.*DELTA; w -= options.dw) { */
    for(w = DELTA; w < 2.*M_PI; w += options.dw) {
	anzelt++;
	wgp = (WG *) malloc(sizeof(WG));
	if (wg == (WG *)NULL) {
	    wg = last = wgp;
	    wg->prevwgp = (WG *)NULL;
	}
	else {
	    wgp->prevwgp = last;
	    last->nextwgp = wgp;
	    last = wgp;
	}
 	last->nextwgp = (WG *)NULL;

	berechnewgpos(w, &(last->wgp));
    }

    if (rechnung_nicht_ok) {

	fprintf (stderr, "Achtung: Probe zeigt Rechenfehler, weiter mit RETURN\n");
#ifndef OS_WIN32
	warten (10000);
#endif
    }
    printwg();
}

void
berechnewgpos(BWINKEL w, WGPOS *wgp) {
    double distanz_p1_e;
    double distanz_p5_i;
    double pdistanz_p1_p3; /* in x/z-Ebene projezierte Strecke p1-p3 */
    double pdistanz_p5_p7; /* in x/y-Ebene projezierte Strecke p5-p7 */
    double ex;
    double ix;
    double s;
    int i;
    double laenge, faktor;
    BWINKEL e1, i1, e5, i5, e51;

    wgp->w = w;

    ex = 2. / sqrt(3.);    /* x-Koordinate des Evolutionspunktes */
    ix = -2. / sqrt(3.);
    s = 1. / sqrt(3.);     /* Lnge einer Scharnierkante */

    /* Die 4 Winkel e1, i1, e5, i5 berechnen.
    ** e1 ist der Winkel zwischen x-Achse und der Strecke p1-e
    ** i1                         x-Achse                 p1-i
    ** e5                         x-Achse                 p5-e
    ** i5                         x-Achse                 p5-i
    */
    if (w > (3./2. *M_PI)) {
	e1 = w - 3./2.*M_PI;
    }
    else if (w > M_PI) {
	e1 = w - 3./2.*M_PI;
    }
    else if (w > M_PI_2) {
	e1 = w - M_PI_2;
    }
    else {
	e1 = w - M_PI_2;
    }
    i1 = w;

    /* e5 liegt vorerst nur im Bereich -PI/2 - +PI/2 und mu noch korrigiert
    ** werden
    */
    e51 = atan (ex * tan(e1) / 1. / sqrt(3.));

    if (w > (3./2.*M_PI)) {
	e5 = M_PI - e51;
	i5 = e5 + M_PI_2;
    }
    else if (w > M_PI) {
	e5 =M_PI - e51;
	i5 = e5 - M_PI_2;
    }
    else if (w > M_PI_2) {
	e5 = 2.*M_PI - e51;
	i5 = e5 + M_PI_2;
    }
    else {
	e5 = 2.*M_PI - e51;
	i5 = e5 - M_PI_2;
    }

/*    fprintf (fw, "xx w=%4.0f e1=%4.0f i1=%4.0f i5=%4.0f e5=%4.0f e51=%4.0f\n", */
/* 	       GWINKEL(w), GWINKEL(e1), GWINKEL(i1), GWINKEL(i5), GWINKEL(e5), */
/* 	       GWINKEL(e51)); */

    /*
    ** Evolutionspunkt
    */
    KP(9,x) = ex;
    KP(9,y) = 0;
    KP(9,z) = ex * tan(e1);

    distanz_p1_e = ex / cos(e1);
    pdistanz_p1_p3 = 1. / distanz_p1_e;

    /*
    ** linker Teil des Wuerfelguertels
    */
    KP(1,x) = 0;
    KP(1,y) = 0;
    KP(1,z) = 0;

    KP(2,x) = s * cos(i1);
    KP(2,y) = 0;
    KP(2,z) = s * sin(i1);

    KP(3,x) = pdistanz_p1_p3 * cos(e1);
    KP(3,y) = sqrt(1. - pow(pdistanz_p1_p3,2.));
    KP(3,z) = pdistanz_p1_p3 * sin(e1);

    KP(4,x) = KP(2,z)*KP(3,y) - KP(2,y)*KP(3,z);
    KP(4,y) = KP(2,x)*KP(3,z) - KP(2,z)*KP(3,x);
    KP(4,z) = KP(2,y)*KP(3,x) - KP(2,x)*KP(3,y);

    KP(4,x) = KP(3,x) - KP(4,x);
    KP(4,y) = KP(3,y) - KP(4,y);
    KP(4,z) = KP(3,z) - KP(4,z);

    /*
    ** w2: Winkel zwischen P1_P3 und der y-Achse
    */
    wgp->w2 = TOBWINKEL(90.) - atan (KP(3,y) / pdistanz_p1_p3);
    if (wgp->w < M_PI) {
	wgp->w2 = -wgp->w2;
    }

    /*
    ** Involutionspunkt
    */
    KP(10,x) = s;
    KP(10,y) = 0.;
    KP(10,z) = s * tan (i1);

    /*   distanz_p5_i = ix / cos(i5); */
    distanz_p5_i = sqrt (pow (KP(10,z), 2.) + pow (2. / sqrt (3.), 2.));
    if (w < M_PI) {
	distanz_p5_i = -distanz_p5_i;
    }
    /* e5 = i5 + M_PI_2; */
 
    /*
    ** rechter Teil des Wuerfelguertels
    ** P5 wird als im 0-Punkt liegend betrachtet
    */
    KP(5,x) = sqrt(3.);
    KP(5,y) = 0.;
    KP(5,z) = 0.;

    /* Die Berechnung von P6 erfolgt hier abweichend von der Beschreibung.
    ** P6 wird so ermittelt, da auf der Geraden P5-E eine Strecke der Lnge
    ** 1/sqrt(3) berechnet wird.
    */
    KP(6,x) = KP(5,x) - KP(9,x);
    KP(6,y) = 0.;
    KP(6,z) = KP(5,z) - KP(9,z);

    laenge = sqrt(pow(KP(6,x),2.) + pow(KP(6,z),2.) + pow(KP(6,y),2.));
    faktor = 1. / (sqrt(3.) * laenge);
    if (w > M_PI) {
	faktor = -faktor;
    }

    KP(6,x) = KP(6,x) * faktor;
    KP(6,y) = 0.;
    KP(6,z) = KP(6,z) * faktor;

    /* Die Berechnung von P7 erfolgt hier abweichend von der Beschreibung.
    ** P7 wird so ermittelt, da auf der Geraden P5-I eine Strecke der Lnge
    ** projezierte Distanz P5-P7 berechnet wird.
    */

    pdistanz_p5_p7 = 1. / distanz_p5_i;

    KP(7,x) = KP(5,x) - KP(10,x);
    KP(7,y) = KP(5,y) - KP(10,y);
    KP(7,z) = KP(5,z) - KP(10,z);

    laenge = sqrt(pow(KP(7,x),2.) + pow(KP(7,z),2.) + pow(KP(7,y),2.));
    faktor = pdistanz_p5_p7 / laenge;
    if (w > M_PI) {
	faktor = -faktor;
    }

    KP(7,x) = KP(7,x) * faktor;
    KP(7,y) = sqrt(1. - pow(pdistanz_p5_p7,2.));
    KP(7,z) = KP(7,z) * faktor;

    KP(8,x) = KP(6,z)*KP(7,y) - KP(6,y)*KP(7,z);
    KP(8,y) = KP(6,x)*KP(7,z) - KP(6,z)*KP(7,x);
    KP(8,z) = KP(6,y)*KP(7,x) - KP(6,x)*KP(7,y);

    KP(8,x) = KP(7,x) - KP(8,x);
    KP(8,y) = KP(7,y) - KP(8,y);
    KP(8,z) = KP(7,z) - KP(8,z);

    /*
    **  den rechten Teil des Wuerfelguertels um sqrt(3.) nach rechts schieben
    */
    /*   KP(5,x) = KP(5,x) + sqrt(3.); */
    KP(6,x) = KP(6,x) + sqrt(3.);
    KP(7,x) = KP(7,x) + sqrt(3.);
    KP(8,x) = KP(8,x) + sqrt(3.);

    /*
    **  unterer Teil des Wuerfelguertels. Spiegelung um die x/z-Ebene, also
    **  y-Koordinate negativ.
    */
    for (i=1; i<=8; i++) {
	KP(i+10,x) =  KP(i,x);
	KP(i+10,y) = -KP(i,y);
	KP(i+10,z) =  KP(i,z);
    }

    /* Auf jeder Koordinatenachse ein Punkt markieren. Die Strecken P1-P19,
    ** P1-P20 und P1-P21 bilden die Koordinatenachsen
    */
    KP(19,x) = 2.;
    KP(19,y) = 0.;
    KP(19,z) = 0.;

    KP(20,x) = 0.;
    KP(20,y) = 2.;
    KP(20,z) = 0.;

    KP(21,x) = 0.;
    KP(21,y) = 0.;
    KP(21,z) = 2.;

    /*
    ** Probe 1: Distanz P3-P7 mu 1 sein
    */
    if ((1. - sqrt (pow (KP(3,x)-KP(7,x), 2.) +
                    pow (KP(3,y)-KP(7,y), 2.) +
                    pow (KP(3,z)-KP(7,z), 2.))) > DELTA) {
	fprintf (stderr, "Distanz P3-P7 bei %7.3f ist %f soll %f\n",
		GWINKEL(w),
		sqrt (pow (KP(3,x)-KP(7,x), 2.) +
		      pow (KP(3,y)-KP(7,y), 2.) +
		      pow (KP(3,z)-KP(7,z), 2.)),
		1.);
	rechnung_nicht_ok = TRUE;
    }
    /*
    ** Probe 2: Distanz P7-P8 mu 1/sqrt(3) sein
    */
    if ((s - sqrt (pow (KP(6,x)-KP(5,x), 2.) +
		   pow (KP(6,y)-KP(5,y), 2.) +
		   pow (KP(6,z)-KP(5,z), 2.))) > DELTA) {
	printf ("Distanz P5-P6 bei %7.3f ist %f soll %f\n",
		GWINKEL(w),
		sqrt (pow (KP(6,x)-KP(5,x), 2.) +
		      pow (KP(6,z)-KP(5,z), 2.) +
		      pow (KP(6,y)-KP(5,y), 2.)),
		s);
	rechnung_nicht_ok = TRUE;
    }
}
